/*
 * Jalview - A Sequence Alignment Editor and Viewer (Version 2.8.2)
 * Copyright (C) 2014 The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.structure;

import jalview.analysis.AlignSeq;
import jalview.api.StructureSelectionManagerProvider;
import jalview.commands.CommandI;
import jalview.commands.EditCommand;
import jalview.commands.EditCommand.Action;
import jalview.commands.EditCommand.Edit;
import jalview.datamodel.AlignedCodonFrame;
import jalview.datamodel.AlignmentAnnotation;
import jalview.datamodel.Annotation;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.SearchResults;
import jalview.datamodel.Sequence;
import jalview.datamodel.SequenceI;
import jalview.io.AppletFormatAdapter;
import jalview.util.MessageManager;
import jalview.util.StringUtils;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import MCview.Atom;
import MCview.PDBChain;

public class StructureSelectionManager
{
  static IdentityHashMap<StructureSelectionManagerProvider, StructureSelectionManager> instances;

  StructureMapping[] mappings;

  private boolean processSecondaryStructure = false;

  private boolean secStructServices = false;

  private boolean addTempFacAnnot = false;

  private Set<AlignedCodonFrame> seqmappings = new LinkedHashSet<AlignedCodonFrame>();

  private List<CommandListener> commandListeners = new ArrayList<CommandListener>();

  /**
   * @return true if will try to use external services for processing secondary
   *         structure
   */
  public boolean isSecStructServices()
  {
    return secStructServices;
  }

  /**
   * control use of external services for processing secondary structure
   * 
   * @param secStructServices
   */
  public void setSecStructServices(boolean secStructServices)
  {
    this.secStructServices = secStructServices;
  }

  /**
   * flag controlling addition of any kind of structural annotation
   * 
   * @return true if temperature factor annotation will be added
   */
  public boolean isAddTempFacAnnot()
  {
    return addTempFacAnnot;
  }

  /**
   * set flag controlling addition of structural annotation
   * 
   * @param addTempFacAnnot
   */
  public void setAddTempFacAnnot(boolean addTempFacAnnot)
  {
    this.addTempFacAnnot = addTempFacAnnot;
  }

  /**
   * 
   * @return if true, the structure manager will attempt to add secondary
   *         structure lines for unannotated sequences
   */

  public boolean isProcessSecondaryStructure()
  {
    return processSecondaryStructure;
  }

  /**
   * Control whether structure manager will try to annotate mapped sequences
   * with secondary structure from PDB data.
   * 
   * @param enable
   */
  public void setProcessSecondaryStructure(boolean enable)
  {
    processSecondaryStructure = enable;
  }

  /**
   * debug function - write all mappings to stdout
   */
  public void reportMapping()
  {
    if (mappings == null)
    {
      System.err.println("reportMapping: No PDB/Sequence mappings.");
    }
    else
    {
      System.err.println("reportMapping: There are " + mappings.length
              + " mappings.");
      for (int m = 0; m < mappings.length; m++)
      {
        System.err.println("mapping " + m + " : " + mappings[m].pdbfile);
      }
    }
  }

  /**
   * map between the PDB IDs (or structure identifiers) used by Jalview and the
   * absolute filenames for PDB data that corresponds to it
   */
  HashMap<String, String> pdbIdFileName = new HashMap<String, String>(),
          pdbFileNameId = new HashMap<String, String>();

  public void registerPDBFile(String idForFile, String absoluteFile)
  {
    pdbIdFileName.put(idForFile, absoluteFile);
    pdbFileNameId.put(absoluteFile, idForFile);
  }

  public String findIdForPDBFile(String idOrFile)
  {
    String id = pdbFileNameId.get(idOrFile);
    return id;
  }

  public String findFileForPDBId(String idOrFile)
  {
    String id = pdbIdFileName.get(idOrFile);
    return id;
  }

  public boolean isPDBFileRegistered(String idOrFile)
  {
    return pdbFileNameId.containsKey(idOrFile)
            || pdbIdFileName.containsKey(idOrFile);
  }

  private static StructureSelectionManager nullProvider = null;

  public static StructureSelectionManager getStructureSelectionManager(
          StructureSelectionManagerProvider context)
  {
    if (context == null)
    {
      if (nullProvider == null)
      {
        if (instances != null)
        {
          throw new Error(MessageManager.getString("error.implementation_error_structure_selection_manager_null"),
                  new NullPointerException(MessageManager.getString("exception.ssm_context_is_null")));
        }
        else
        {
          nullProvider = new StructureSelectionManager();
        }
        return nullProvider;
      }
    }
    if (instances == null)
    {
      instances = new java.util.IdentityHashMap<StructureSelectionManagerProvider, StructureSelectionManager>();
    }
    StructureSelectionManager instance = instances.get(context);
    if (instance == null)
    {
      if (nullProvider != null)
      {
        instance = nullProvider;
      }
      else
      {
        instance = new StructureSelectionManager();
      }
      instances.put(context, instance);
    }
    return instance;
  }

  /**
   * flag controlling whether SeqMappings are relayed from received sequence
   * mouse over events to other sequences
   */
  boolean relaySeqMappings = true;

  /**
   * Enable or disable relay of seqMapping events to other sequences. You might
   * want to do this if there are many sequence mappings and the host computer
   * is slow
   * 
   * @param relay
   */
  public void setRelaySeqMappings(boolean relay)
  {
    relaySeqMappings = relay;
  }

  /**
   * get the state of the relay seqMappings flag.
   * 
   * @return true if sequence mouse overs are being relayed to other mapped
   *         sequences
   */
  public boolean isRelaySeqMappingsEnabled()
  {
    return relaySeqMappings;
  }

  Vector listeners = new Vector();

  /**
   * register a listener for alignment sequence mouseover events
   * 
   * @param svl
   */
  public void addStructureViewerListener(Object svl)
  {
    if (!listeners.contains(svl))
    {
      listeners.addElement(svl);
    }
  }

  public String alreadyMappedToFile(String pdbid)
  {
    if (mappings != null)
    {
      for (int i = 0; i < mappings.length; i++)
      {
        if (mappings[i].getPdbId().equals(pdbid))
        {
          return mappings[i].pdbfile;
        }
      }
    }
    return null;
  }

  /**
   * Import structure data and register a structure mapping for broadcasting
   * colouring, mouseovers and selection events (convenience wrapper).
   * 
   * @param sequence
   *          - one or more sequences to be mapped to pdbFile
   * @param targetChains
   *          - optional chain specification for mapping each sequence to pdb
   *          (may be nill, individual elements may be nill)
   * @param pdbFile
   *          - structure data resource
   * @param protocol
   *          - how to resolve data from resource
   * @return null or the structure data parsed as a pdb file
   */
  synchronized public MCview.PDBfile setMapping(SequenceI[] sequence,
          String[] targetChains, String pdbFile, String protocol)
  {
    return setMapping(true, sequence, targetChains, pdbFile, protocol);
  }

  /**
   * create sequence structure mappings between each sequence and the given
   * pdbFile (retrieved via the given protocol).
   * 
   * @param forStructureView
   *          when true, record the mapping for use in mouseOvers
   * 
   * @param sequence
   *          - one or more sequences to be mapped to pdbFile
   * @param targetChains
   *          - optional chain specification for mapping each sequence to pdb
   *          (may be nill, individual elements may be nill)
   * @param pdbFile
   *          - structure data resource
   * @param protocol
   *          - how to resolve data from resource
   * @return null or the structure data parsed as a pdb file
   */
  synchronized public MCview.PDBfile setMapping(boolean forStructureView,
          SequenceI[] sequence,
          String[] targetChains, String pdbFile, String protocol)
  {
    /*
     * There will be better ways of doing this in the future, for now we'll use
     * the tried and tested MCview pdb mapping
     */
    MCview.PDBfile pdb = null;
    boolean parseSecStr = processSecondaryStructure;
    if (isPDBFileRegistered(pdbFile))
    {
      for (SequenceI sq : sequence)
      {
        SequenceI ds = sq;
        while (ds.getDatasetSequence() != null)
        {
          ds = ds.getDatasetSequence();
        }
        ;
        if (ds.getAnnotation() != null)
        {
          for (AlignmentAnnotation ala : ds.getAnnotation())
          {
            // false if any annotation present from this structure
            // JBPNote this fails for jmol/chimera view because the *file* is
            // passed, not the structure data ID -
            if (MCview.PDBfile.isCalcIdForFile(ala,
                    findIdForPDBFile(pdbFile)))
            {
              parseSecStr = false;
            }
          }
        }
      }
    }
    try
    {
      pdb = new MCview.PDBfile(addTempFacAnnot, parseSecStr,
              secStructServices, pdbFile, protocol);
      if (pdb.id != null && pdb.id.trim().length() > 0
              && AppletFormatAdapter.FILE.equals(protocol))
      {
        registerPDBFile(pdb.id.trim(), pdbFile);
      }
    } catch (Exception ex)
    {
      ex.printStackTrace();
      return null;
    }

    String targetChain;
    for (int s = 0; s < sequence.length; s++)
    {
      boolean infChain = true;
      if (targetChains != null && targetChains[s] != null)
      {
        infChain = false;
        targetChain = targetChains[s];
      }
      else if (sequence[s].getName().indexOf("|") > -1)
      {
        targetChain = sequence[s].getName().substring(
                sequence[s].getName().lastIndexOf("|") + 1);
        if (targetChain.length() > 1)
        {
          if (targetChain.trim().length() == 0)
          {
            targetChain = " ";
          }
          else
          {
            // not a valid chain identifier
            targetChain = "";
          }
        }
      }
      else
      {
        targetChain = "";
      }

      int max = -10;
      AlignSeq maxAlignseq = null;
      String maxChainId = " ";
      PDBChain maxChain = null;
      boolean first = true;
      for (int i = 0; i < pdb.chains.size(); i++)
      {
        PDBChain chain = (pdb.chains.elementAt(i));
        if (targetChain.length() > 0 && !targetChain.equals(chain.id)
                && !infChain)
        {
          continue; // don't try to map chains don't match.
        }
        // TODO: correctly determine sequence type for mixed na/peptide
        // structures
        AlignSeq as = new AlignSeq(sequence[s],
                pdb.chains.elementAt(i).sequence,
                pdb.chains.elementAt(i).isNa ? AlignSeq.DNA
                        : AlignSeq.PEP);
        as.calcScoreMatrix();
        as.traceAlignment();

        if (first || as.maxscore > max
                || (as.maxscore == max && chain.id.equals(targetChain)))
        {
          first = false;
          maxChain = chain;
          max = as.maxscore;
          maxAlignseq = as;
          maxChainId = chain.id;
        }
      }
      if (maxChain == null)
      {
        continue;
      }
      final StringBuffer mappingDetails = new StringBuffer();
      mappingDetails.append("\n\nPDB Sequence is :\nSequence = "
              + maxChain.sequence.getSequenceAsString());
      mappingDetails.append("\nNo of residues = "
              + maxChain.residues.size() + "\n\n");
      PrintStream ps = new PrintStream(System.out)
      {
        @Override
        public void print(String x)
        {
          mappingDetails.append(x);
        }

        @Override
        public void println()
        {
          mappingDetails.append("\n");
        }
      };

      maxAlignseq.printAlignment(ps);

      mappingDetails.append("\nPDB start/end " + maxAlignseq.seq2start
              + " " + maxAlignseq.seq2end);
      mappingDetails.append("\nSEQ start/end "
              + (maxAlignseq.seq1start + sequence[s].getStart() - 1) + " "
              + (maxAlignseq.seq1end + sequence[s].getEnd() - 1));

      maxChain.makeExactMapping(maxAlignseq, sequence[s]);
      jalview.datamodel.Mapping sqmpping = maxAlignseq
              .getMappingFromS1(false);
      jalview.datamodel.Mapping omap = new jalview.datamodel.Mapping(
              sqmpping.getMap().getInverse());
      maxChain.transferRESNUMFeatures(sequence[s], null);

      // allocate enough slots to store the mapping from positions in
      // sequence[s] to the associated chain
      int[][] mapping = new int[sequence[s].findPosition(sequence[s]
              .getLength()) + 2][2];
      int resNum = -10000;
      int index = 0;

      do
      {
        Atom tmp = (Atom) maxChain.atoms.elementAt(index);
        if (resNum != tmp.resNumber && tmp.alignmentMapping != -1)
        {
          resNum = tmp.resNumber;
          mapping[tmp.alignmentMapping + 1][0] = tmp.resNumber;
          mapping[tmp.alignmentMapping + 1][1] = tmp.atomIndex;
        }

        index++;
      } while (index < maxChain.atoms.size());

      if (protocol.equals(jalview.io.AppletFormatAdapter.PASTE))
      {
        pdbFile = "INLINE" + pdb.id;
      }
      StructureMapping newMapping = new StructureMapping(sequence[s],
              pdbFile, pdb.id, maxChainId, mapping,
              mappingDetails.toString());
      if (forStructureView)
      {

        if (mappings == null)
        {
          mappings = new StructureMapping[1];
        }
        else
        {
          StructureMapping[] tmp = new StructureMapping[mappings.length + 1];
          System.arraycopy(mappings, 0, tmp, 0, mappings.length);
          mappings = tmp;
        }

        mappings[mappings.length - 1] = newMapping;
      }
      maxChain.transferResidueAnnotation(newMapping, sqmpping);
    }
    // ///////

    return pdb;
  }

  public void removeStructureViewerListener(Object svl, String[] pdbfiles)
  {
    listeners.removeElement(svl);
    if (svl instanceof SequenceListener)
    {
      for (int i = 0; i < listeners.size(); i++)
      {
        if (listeners.elementAt(i) instanceof StructureListener)
        {
          ((StructureListener) listeners.elementAt(i))
                  .releaseReferences(svl);
        }
      }
    }

    if (pdbfiles == null)
    {
      return;
    }
    String[] handlepdbs;
    Vector pdbs = new Vector();
    for (int i = 0; i < pdbfiles.length; pdbs.addElement(pdbfiles[i++]))
    {
      ;
    }
    StructureListener sl;
    for (int i = 0; i < listeners.size(); i++)
    {
      if (listeners.elementAt(i) instanceof StructureListener)
      {
        sl = (StructureListener) listeners.elementAt(i);
        handlepdbs = sl.getPdbFile();
        for (int j = 0; j < handlepdbs.length; j++)
        {
          if (pdbs.contains(handlepdbs[j]))
          {
            pdbs.removeElement(handlepdbs[j]);
          }
        }

      }
    }

    if (pdbs.size() > 0 && mappings != null)
    {
      Vector tmp = new Vector();
      for (int i = 0; i < mappings.length; i++)
      {
        if (!pdbs.contains(mappings[i].pdbfile))
        {
          tmp.addElement(mappings[i]);
        }
      }

      mappings = new StructureMapping[tmp.size()];
      tmp.copyInto(mappings);
    }
  }

  public void mouseOverStructure(int pdbResNum, String chain, String pdbfile)
  {
    if (listeners == null)
    {
      // old or prematurely sent event
      return;
    }
    boolean hasSequenceListeners = handlingVamsasMo || seqmappings != null;
    SearchResults results = null;
    SequenceI lastseq = null;
    int lastipos = -1, indexpos;
    for (int i = 0; i < listeners.size(); i++)
    {
      if (listeners.elementAt(i) instanceof SequenceListener)
      {
        if (results == null)
        {
          results = new SearchResults();
        }
        if (mappings != null)
        {
          for (int j = 0; j < mappings.length; j++)
          {
            if (mappings[j].pdbfile.equals(pdbfile)
                    && mappings[j].pdbchain.equals(chain))
            {
              indexpos = mappings[j].getSeqPos(pdbResNum);
              if (lastipos != indexpos && lastseq != mappings[j].sequence)
              {
                results.addResult(mappings[j].sequence, indexpos, indexpos);
                lastipos = indexpos;
                lastseq = mappings[j].sequence;
                // construct highlighted sequence list
                if (seqmappings != null)
                {
                  for (AlignedCodonFrame acf : seqmappings)
                  {
                    acf.markMappedRegion(mappings[j].sequence, indexpos,
                            results);
                  }
                }
              }
            }
          }
        }
      }
    }
    if (results != null)
    {
      for (int i = 0; i < listeners.size(); i++)
      {
        Object li = listeners.elementAt(i);
        if (li instanceof SequenceListener)
        {
          ((SequenceListener) li).highlightSequence(results);
        }
      }
    }
  }

  /**
   * highlight regions associated with a position (indexpos) in seq
   * 
   * @param seq
   *          the sequence that the mouse over occurred on
   * @param indexpos
   *          the absolute position being mouseovered in seq (0 to seq.length())
   * @param index
   *          the sequence position (if -1, seq.findPosition is called to
   *          resolve the residue number)
   */
  public void mouseOverSequence(SequenceI seq, int indexpos, int index,
          VamsasSource source)
  {
    boolean hasSequenceListeners = handlingVamsasMo || seqmappings != null;
    SearchResults results = null;
    if (index == -1)
    {
      index = seq.findPosition(indexpos);
    }
    for (int i = 0; i < listeners.size(); i++)
    {
      Object listener = listeners.elementAt(i);
      if (listener == source)
      {
        // TODO listener (e.g. SeqPanel) is never == source (AlignViewport)
        // Temporary fudge with SequenceListener.getVamsasSource()
        continue;
      }
      if (listener instanceof StructureListener)
      {
        highlightStructure((StructureListener) listener, seq, index);
      }
      else
      {
        if (listener instanceof SequenceListener)
        {
          final SequenceListener seqListener = (SequenceListener) listener;
          if (hasSequenceListeners
                  && seqListener.getVamsasSource() != source)
          {
            if (relaySeqMappings)
            {
              if (results == null)
              {
                results = buildSearchResults(seq, index);
              }
              seqListener.highlightSequence(results);
            }
          }
        }
        else if (listener instanceof VamsasListener && !handlingVamsasMo)
        {
          ((VamsasListener) listener).mouseOverSequence(seq, indexpos,
                  source);
        }
        else if (listener instanceof SecondaryStructureListener)
        {
          ((SecondaryStructureListener) listener).mouseOverSequence(seq,
                  indexpos);
        }
      }
    }
  }

  /**
   * Returns a SearchResults object describing the mapped region corresponding
   * to the specified sequence position.
   * 
   * @param seq
   * @param index
   * @return
   */
  protected SearchResults buildSearchResults(SequenceI seq, int index)
  {
    SearchResults results;
    results = new SearchResults();
    if (index >= seq.getStart() && index <= seq.getEnd())
    {
      if (seqmappings != null)
      {
        for (AlignedCodonFrame acf : seqmappings)
        {
          acf.markMappedRegion(seq, index, results);
        }
      }
      // hasSequenceListeners = results.getSize() > 0;
      if (handlingVamsasMo)
      {
        // maybe have to resolve seq to a dataset sequence...
        // add in additional direct sequence and/or dataset sequence
        // highlighting
        results.addResult(seq, index, index);
      }
    }
    return results;
  }

  /**
   * Send suitable messages to a StructureListener to highlight atoms
   * corresponding to the given sequence position.
   * 
   * @param sl
   * @param seq
   * @param index
   */
  protected void highlightStructure(StructureListener sl, SequenceI seq,
          int index)
  {
    int atomNo;
    if (mappings != null)
    {
      List<AtomSpec> atoms = new ArrayList<AtomSpec>();
      for (int j = 0; j < mappings.length; j++)
      {
        if (mappings[j].sequence == seq
                || mappings[j].sequence == seq.getDatasetSequence())
        {
          atomNo = mappings[j].getAtomNum(index);

          if (atomNo > 0)
          {
            atoms.add(new AtomSpec(mappings[j].pdbfile,
                    mappings[j].pdbchain, mappings[j].getPDBResNum(index),
                    atomNo));
          }
        }
      }
      sl.highlightAtoms(atoms);
    }
  }

  /**
   * true if a mouse over event from an external (ie Vamsas) source is being
   * handled
   */
  boolean handlingVamsasMo = false;

  long lastmsg = 0;

  /**
   * as mouseOverSequence but only route event to SequenceListeners
   * 
   * @param sequenceI
   * @param position
   *          in an alignment sequence
   */
  public void mouseOverVamsasSequence(SequenceI sequenceI, int position,
          VamsasSource source)
  {
    handlingVamsasMo = true;
    long msg = sequenceI.hashCode() * (1 + position);
    if (lastmsg != msg)
    {
      lastmsg = msg;
      mouseOverSequence(sequenceI, position, -1, source);
    }
    handlingVamsasMo = false;
  }

  public Annotation[] colourSequenceFromStructure(SequenceI seq,
          String pdbid)
  {
    return null;
    // THIS WILL NOT BE AVAILABLE IN JALVIEW 2.3,
    // UNTIL THE COLOUR BY ANNOTATION IS REWORKED
    /*
     * Annotation [] annotations = new Annotation[seq.getLength()];
     * 
     * StructureListener sl; int atomNo = 0; for (int i = 0; i <
     * listeners.size(); i++) { if (listeners.elementAt(i) instanceof
     * StructureListener) { sl = (StructureListener) listeners.elementAt(i);
     * 
     * for (int j = 0; j < mappings.length; j++) {
     * 
     * if (mappings[j].sequence == seq && mappings[j].getPdbId().equals(pdbid)
     * && mappings[j].pdbfile.equals(sl.getPdbFile())) {
     * System.out.println(pdbid+" "+mappings[j].getPdbId() +"
     * "+mappings[j].pdbfile);
     * 
     * java.awt.Color col; for(int index=0; index<seq.getLength(); index++) {
     * if(jalview.util.Comparison.isGap(seq.getCharAt(index))) continue;
     * 
     * atomNo = mappings[j].getAtomNum(seq.findPosition(index)); col =
     * java.awt.Color.white; if (atomNo > 0) { col = sl.getColour(atomNo,
     * mappings[j].getPDBResNum(index), mappings[j].pdbchain,
     * mappings[j].pdbfile); }
     * 
     * annotations[index] = new Annotation("X",null,' ',0,col); } return
     * annotations; } } } }
     * 
     * return annotations;
     */
  }

  public void structureSelectionChanged()
  {
  }

  public void sequenceSelectionChanged()
  {
  }

  public void sequenceColoursChanged(Object source)
  {
    StructureListener sl;
    for (int i = 0; i < listeners.size(); i++)
    {
      if (listeners.elementAt(i) instanceof StructureListener)
      {
        sl = (StructureListener) listeners.elementAt(i);
        sl.updateColours(source);
      }
    }
  }

  public StructureMapping[] getMapping(String pdbfile)
  {
    Vector tmp = new Vector();
    if (mappings != null)
    {
      for (int i = 0; i < mappings.length; i++)
      {
        if (mappings[i].pdbfile.equals(pdbfile))
        {
          tmp.addElement(mappings[i]);
        }
      }
    }
    StructureMapping[] ret = new StructureMapping[tmp.size()];
    for (int i = 0; i < tmp.size(); i++)
    {
      ret[i] = (StructureMapping) tmp.elementAt(i);
    }

    return ret;
  }

  public String printMapping(String pdbfile)
  {
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < mappings.length; i++)
    {
      if (mappings[i].pdbfile.equals(pdbfile))
      {
        sb.append(mappings[i].mappingDetails);
      }
    }

    return sb.toString();
  }

  /**
   * Remove each of the given codonFrames from the stored set (if present).
   * 
   * @param codonFrames
   */
  public void removeMappings(AlignedCodonFrame[] codonFrames)
  {
    if (codonFrames != null)
    {
      for (AlignedCodonFrame acf : codonFrames)
      {
        seqmappings.remove(acf);
      }
    }
  }

  /**
   * Add each of the given codonFrames to the stored set (if not aready
   * present).
   * 
   * @param codonFrames
   */
  public void addMappings(AlignedCodonFrame[] codonFrames)
  {
    if (codonFrames != null)
    {
      for (AlignedCodonFrame acf : codonFrames)
      {
        seqmappings.add(acf);
      }
    }
  }

  Vector<SelectionListener> sel_listeners = new Vector<SelectionListener>();

  public void addSelectionListener(SelectionListener selecter)
  {
    if (!sel_listeners.contains(selecter))
    {
      sel_listeners.addElement(selecter);
    }
  }

  public void removeSelectionListener(SelectionListener toremove)
  {
    if (sel_listeners.contains(toremove))
    {
      sel_listeners.removeElement(toremove);
    }
  }

  public synchronized void sendSelection(
          jalview.datamodel.SequenceGroup selection,
          jalview.datamodel.ColumnSelection colsel, SelectionSource source)
  {
    if (sel_listeners != null && sel_listeners.size() > 0)
    {
      Enumeration listeners = sel_listeners.elements();
      while (listeners.hasMoreElements())
      {
        SelectionListener slis = ((SelectionListener) listeners
                .nextElement());
        if (slis != source)
        {
          slis.selection(selection, colsel, source);
        }
        ;
      }
    }
  }

  Vector<AlignmentViewPanelListener> view_listeners = new Vector<AlignmentViewPanelListener>();

  public synchronized void sendViewPosition(
          jalview.api.AlignmentViewPanel source, int startRes, int endRes,
          int startSeq, int endSeq)
  {

    if (view_listeners != null && view_listeners.size() > 0)
    {
      Enumeration<AlignmentViewPanelListener> listeners = view_listeners
              .elements();
      while (listeners.hasMoreElements())
      {
        AlignmentViewPanelListener slis = listeners.nextElement();
        if (slis != source)
        {
          slis.viewPosition(startRes, endRes, startSeq, endSeq, source);
        }
        ;
      }
    }
  }

  /**
   * release all references associated with this manager provider
   * 
   * @param jalviewLite
   */
  public static void release(StructureSelectionManagerProvider jalviewLite)
  {
    // synchronized (instances)
    {
      if (instances == null)
      {
        return;
      }
      StructureSelectionManager mnger = (instances.get(jalviewLite));
      if (mnger != null)
      {
        instances.remove(jalviewLite);
        try
        {
          mnger.finalize();
        } catch (Throwable x)
        {
        }
      }
    }
  }

  public void registerPDBEntry(PDBEntry pdbentry)
  {
    if (pdbentry.getFile() != null
            && pdbentry.getFile().trim().length() > 0)
    {
      registerPDBFile(pdbentry.getId(), pdbentry.getFile());
    }
  }

  public void addCommandListener(CommandListener cl)
  {
    if (!commandListeners.contains(cl))
    {
      commandListeners.add(cl);
    }
  }

  public boolean hasCommandListener(CommandListener cl)
  {
    return this.commandListeners.contains(cl);
  }

  public boolean removeEditListener(CommandListener l)
  {
    return commandListeners.remove(l);
  }

  /**
   * Forward a command to any command listeners (except for the command's
   * source).
   * 
   * @param command
   *          the command to be broadcast (in its form after being performed)
   * @param undo
   *          if true, the command was being 'undone'
   * @param source
   */
  public void commandPerformed(CommandI command, boolean undo,
          VamsasSource source)
  {
    for (CommandListener listener : commandListeners)
    {
      if (listener.getVamsasSource() != source)
      {
        listener.mirrorCommand(command, undo, this);
      }
    }
  }

  /**
   * Returns a new EditCommand representing the given command as mapped to the
   * given sequences. If there is no mapping, returns an empty EditCommand.
   * 
   * @param command
   * @param undo
   * @param targetSeqs
   * @param gapChar
   * @return
   */
  public EditCommand mapEditCommand(EditCommand command, boolean undo,
          final List<SequenceI> targetSeqs, char gapChar)
  {
    /*
     * Cache a copy of the target sequences so we can mimic successive edits on
     * them. This lets us compute mappings for all edits in the set.
     */
    Map<SequenceI, SequenceI> targetCopies = new HashMap<SequenceI, SequenceI>();
    for (SequenceI seq : targetSeqs)
    {
      SequenceI ds = seq.getDatasetSequence();
      if (ds != null)
      {
        final Sequence copy = new Sequence("", new String(seq.getSequence()));
        copy.setDatasetSequence(ds);
        targetCopies.put(ds, copy);
      }
    }

    /*
     * Compute 'source' sequences as they were before applying edits:
     */
    Map<SequenceI, SequenceI> originalSequences = command.priorState(undo);

    EditCommand result = new EditCommand();
    Iterator<Edit> edits = command.getEditIterator(!undo);
    while (edits.hasNext())
    {
      Edit edit = edits.next();
      Action action = edit.getAction();

      /*
       * Invert sense of action if an Undo.
       */
      if (undo)
      {
        action = action == Action.INSERT_GAP ? Action.DELETE_GAP
                : (action == Action.DELETE_GAP ? Action.INSERT_GAP : action);
      }
      final int count = edit.getNumber();
      final int editPos = edit.getPosition();
      for (SequenceI seq : edit.getSequences())
      {
        /*
         * Get residue position at (or to right of) edit location. Note we use
         * our 'copy' of the sequence before editing for this.
         */
        SequenceI ds = seq.getDatasetSequence();
        if (ds == null)
        {
          continue;
        }
        final SequenceI actedOn = originalSequences.get(ds);
        final int seqpos = actedOn.findPosition(editPos);

        /*
         * Determine all mappings from this position to mapped sequences.
         */
        SearchResults sr = buildSearchResults(seq, seqpos);

        if (!sr.isEmpty())
        {
          for (SequenceI targetSeq : targetSeqs)
          {
            ds = targetSeq.getDatasetSequence();
            if (ds == null)
            {
              continue;
            }
            SequenceI copyTarget = targetCopies.get(ds);
            final int[] match = sr.getResults(copyTarget, 0,
                    copyTarget.getLength());
            if (match != null)
            {
              final int ratio = 3; // TODO: compute this - how?
              final int mappedCount = count * ratio;

              /*
               * Shift Delete start position left, as it acts on positions to
               * its right.
               */
              int mappedEditPos = action == Action.DELETE_GAP ? match[0]
                      - mappedCount : match[0];
              Edit e = new EditCommand().new Edit(action, new SequenceI[]
              { targetSeq }, mappedEditPos, mappedCount, gapChar);
              result.addEdit(e);

              /*
               * and 'apply' the edit to our copy of its target sequence
               */
              if (action == Action.INSERT_GAP)
              {
                copyTarget.setSequence(new String(StringUtils.insertCharAt(
                        copyTarget.getSequence(), mappedEditPos,
                        mappedCount,
                        gapChar)));
              }
              else if (action == Action.DELETE_GAP)
              {
                copyTarget.setSequence(new String(StringUtils.deleteChars(
                        copyTarget.getSequence(), mappedEditPos,
                        mappedEditPos + mappedCount)));
              }
            }
          }
        }
        /*
         * and 'apply' the edit to our copy of its source sequence
         */
        if (action == Action.INSERT_GAP) {
          actedOn.setSequence(new String(StringUtils.insertCharAt(
                  actedOn.getSequence(), editPos, count, gapChar)));
        }
        else if (action == Action.DELETE_GAP)
        {
          actedOn.setSequence(new String(StringUtils.deleteChars(
                  actedOn.getSequence(), editPos, editPos + count)));
        }
      }
    }
    return result;
  }
}
